#!/usr/bin/ruby
#
# Script version 1.8
# Supported Toolpack versions: 2.2, 2.6, 2.7 and higher
#
# Version history:
#  1.0  First draft
#  1.1  Added pkg_path autodiscovery, or parameter from YML config file, so that
#        source code does not need modification anymore to run on different 
#        release
#       PATH and LD_LIBRARY env vars are being set dynamically to find 
#        required DLL for toolpack.  So no need to copy dll on windows anymore.
#       Fixed memory leakage introduced when using Toolpack release 2.6
#       WriteLock, For both CSV and SQLITE files, now we have a max of 5 retries, 
#        with 1 second of sleep between each, to wait for unlocking.  
#        If file is not unlocked in time, update data is lost.
#  1.2  Disable vacuuming as sqlite database does not need to shrink,
#        as the number of entries fixed by the 'rotation_period', 
#        thus a stable file size.  Vacuuming takes a lot of DISK/CPU power, 
#        and is now done at application startup only
#       Fixed TimerPeriod so that event triggers on localtime periods
#       Will print value 'N/A' for invalid/non-existing statistic elements on a path
#       Now working on Toolpack 2.2
#  1.3  Removed dependance to sqlite3 when not being used.
#       Fixed printing of unexisting subfield (ie: stats.subfield)
#       Added support for extended fields ('x') and Level 0 fields
#       Added :dump command line argument to dump all paths
#       Added config file validation against "paths" without "fields"
#  1.4  Fixed subfield defintion with number. ie: mtp2_statistic_struct.ack_failure_cnt
#       Added string representation for slog_update_interval (s,m,h,d) 
#       and slogs_size_max_mb (K,M,G) in configuration file
#  1.5  Add field query on command line (ex. /nap:SIP_NAP::asr_stats_incoming_struct.last_24h_call_cnt)
#       Support query of field stat name with space and parentheses in .yml config file
#         ex. remote_drop_stats.DESTINATION_OUT_OF_ORDER (27)
#  1.6  Fix Tracking 19934 - wrong package directory returned
#       Fix Tracking 19935 - failed with uninitialized constant StatusClientWrapper::StatusClient on 2.9
#  1.7  Fix Tracking 21349 - Adjusted tbstatus.rb to support toolpack version 2.9
#  1.8  Fix small issue in 1.7  affecting tbstatus only when executed with
#        a group. i.e tbstatus.rb [GROUP]

USAGE = <<-USAGE
Usage: tbstatus.rb [path]
  path: '/*', /nap or any other supported path

Usage: tbstatus.rb [GROUP]
  GROUP: any defined GROUP in yml config file in CAPITAL letters
         (:all to print them all)

Usage: tbstatus.rb [command]
  command: :all to print statistics for all defined GROUPS
           :dump to print all statistics for all supported paths

Usage: tbstatus.rb -d
  Goes into deamon mode, and start logging system statistics
  according to YML config file

CFG: -f can be used as first argument to specify YML cfg file

USAGE

require 'rubygems'
require 'yaml'
require 'erb'
require 'logger'

PORTAL_PATH                 = ""            # Dynamically set later
CLEANUP_PERIOD              = 60 * 60
DEFAULT_SLOG_UPDATE_INTERVAL= '5m'          # 5 minutes
DEFAULT_SLOG_SIZE_MAX_MB    = '10M'         # 10 megabytes
FILE_WRITELOCK_RETRY_CNT    = 5
DEFAULT_YML_CFG_FILE        = 'tbstatus.yml'

#
# OrderedHash (missing in pre-2.7 release)
class OrderedHash < Hash
  def initialize
    @keys = []
    @array = @keys # for 2.6 back compatibility
  end
  def []=( key, value )
    @keys << key unless @keys.include? key
    super
  end
  def each
    @keys.each do |key|
      yield key, self[key]
    end
  end
  def at(index)
    self[@keys[index]]
  end
  def keys
    @keys
  end
  def collect
    @keys.collect { |key| yield key, self[key] }
  end
end

# Converts period string (1s,1m,1h,1d) into number of seconds
class SecondInterval
  def self.parse(interval)
    @interval = case interval.to_s
      when /^(\d+)$/     : $1.to_i  # default unit is seconds
      when /^(\d+)[sS]$/ : $1.to_i
      when /^(\d+)[mM]$/ : $1.to_i * 60
      when /^(\d+)[hH]$/ : $1.to_i * 60 * 60
      when /^(\d+)[dD]$/ : $1.to_i * 60 * 60 * 24
      else
        interval.to_i
      end
    #puts "SecondInterval: #{interval} => #{@interval}"
    @interval
  end
end

# Converts filesize string (1s,1m,1h,1d) into number of seconds
class FileSize
  def self.parse(size)
    @size = case size.to_s
      when /^(\d+)$/     : $1.to_i * 1024 * 1024  # default unit is megabytes
      when /^(\d+)[kK]$/ : $1.to_i * 1024
      when /^(\d+)[mM]$/ : $1.to_i * 1024 * 1024
      when /^(\d+)[gG]$/ : $1.to_i * 1024 * 1024 * 1024
      else
        size.to_i * 1024
      end
    #puts "FileSize: #{size} => #{@size}"
    @size
  end
end

class PeriodTimer
  # period_time:
  #  Period time in seconds. Ie, 15*60 would mean begining of 
  #  every 15 minutes period in an hour (0,15,30,45)
  # period_distance:
  #  Distance from current period.  0 would mean now, 1 would mean begining of next period.
  SECONDS_IN_DAY = 24 * 60 * 60
  def initialize(period_time, period_distance)
    @gmc_offset = (period_time%(SECONDS_IN_DAY)) == 0 ?  Time.now.gmt_offset : 0 # if aligned on a day, then adjust event to 00:00 locatime
    @period_time = period_time
    @next_period_time = ((Time.now.to_i / @period_time) + period_distance ) * @period_time # At begining of every period
  end
  def now?
    (Time.now.to_i + @gmc_offset) >= @next_period_time  # Time.now.to_i is the UTC time, so we need to adjust with gmc offset
  end
  def shift(period_distance=1)
    @next_period_time += (@period_time*period_distance)
  end
end

#
# Toolpack StatusClient wrapper
class StatusClientWrapper
  attr_reader :toolpack_version

  def initialize(params = {})
    adjust_env(params)
    @status_client = init_status(params[:gw_port] || ENV['TBX_GW_PORT'])
  end

  def uninitialize
    uninit_status
  end

  # Prints given path
  # path:
  #   Stats path to print
  # fields_filter:
  #   Filters the printing of only desired fields
  def print_path(path, fields_filter = nil)
    status = get path
    print_status status, path, fields_filter
    garbage_collect
  end

  def get(path)
    status = @status_client.get path, { :options => 'x', :level => '0' }
  end

  # StatusClient.release method was removed in 2.7
  def garbage_collect
    @status_client.garbage_collect
  end

  # Print the field node or the field_list node.
  # field:
  #   Field node to print
  # subfield_name:
  #   Filters the printing of only desired subfield
  def print_field(field, indent = 0, subfield_name=nil)
    str = "  - #{field.get_name}".ljust(50-indent) + " #{field.to_s}"
    puts "".ljust(indent) + str
    indent += 4
    case field.get_type
    when :list  #field node might be a list of subfields
      if subfield_name.nil?
        field.get_fields.each { |k,f| print_field(f,indent) }
      else
        unless field.get_fields[subfield_name].nil?
          print_field field.get_fields[subfield_name],indent
        else
          # The fieldname does not exist, in this case we print N/A
          str = "  - #{subfield_name}".ljust(50-indent) + " N/A"
          puts "".ljust(indent) + str
        end
      end
    end
  end

  # Print the status node and all of its fields.
  # status:
  #  Status node to print
  # path:
  #  Current path, used to know if we need to print childs
  # fields_filter: 
  #  Filters the printing of only desired fields
  def print_status(status, path, fields_filter = nil)
    status.each do |name,stats|
      puts ""
      puts "#{stats.get_path}"
      puts "#{name} - #{stats.get_description}"
      if fields_filter.nil?
        stats.get_fields.each do |name,stat_field|
          print_field stat_field
        end
      else
        fields_filter.each do |field_name|
          if field_name =~ /^([a-zA-Z_0-9]*)\.([a-zA-Z_0-9 \(\)]*)$/
            stat_field = case
              when toolpack_version >= 209000 
                stats[$1] rescue nil # test stats if supports the requested field
              else
                stats.send($1) rescue nil # test stats if supports the requested field
              end
            unless stat_field.nil?
              print_field stat_field, 0, $2 
            else
              # The fieldname does not exist, in this case we print N/A
              puts "  - #{$1}".ljust(50) + " N/A"
            end
          else
            stat_field = case
              when toolpack_version >= 209000 
                stats[field_name] rescue nil # test stats if supports the requested field
              else
                stats.send(field_name) rescue nil # test stats if supports the requested field
              end
            unless stat_field.nil?
              print_field stat_field
            else
              # The fieldname does not exist, in this case we print N/A
              puts "  - #{field_name}".ljust(50) + " N/A"
            end
    	  end
        end
      end
      puts ""
      print_status( get_child(stats), path ) if path =~ /.*\*$/ 
    end
  end
  def get_child(parent)
    begin
      parent.get_child('/*')
    rescue
      # go around 2.2 bug
      children_stats = {}
      children = parent.get_stat.GetChildStat('/*')
      (0..(children.Length-1)).each do |idx|
        child = children.GetItem(idx)
        children_stats[ child.GetName ] = StatusElement.new(child)
      end 
      children_stats
    end
  end

  # Get the list of available paths on which to query stats
  def get_paths
    @status_client.get_paths
  end
private
  # Initialize Toolpack Status Client
  def init_status gw_port  
    silence_stdout do
      @toolpack_version = get_toopack_version( @toopack_pkg_path )
      require "toolpack" unless @pre_2_6
      require "status_base" rescue nil #rescue to get around load problem in 2.2
      require "status_client" unless @toolpack_version < 209000
      status_client = StatusClient.new :gw_port => gw_port.to_i
      sleep 2 if @pre_2_6  # Give some time to learn available paths
      status_client
    end
  end

  # Uninitialize Toolpack Status Client
  def uninit_status
    silence_stdout do
      StatusClient.stop_client
    end
  end

  #
  # When initilasing StatusClient, the toolapck DLL prints
  # all sorts of garbage on stdout, so we need to filter them out
  def silence_stdout
    nulldevice = RUBY_PLATFORM =~ /mswin|mingw/ ? 'NUL:' : '/dev/null'
    orig_std_out = STDOUT.clone
    STDOUT.reopen(File.open(nulldevice, 'w+'))
    yield if block_given?
  ensure
    STDOUT.reopen(orig_std_out) if block_given?
  end
  def adjust_env(params)
    # Get toolpack pkg_path from YML config file, or find latest installed one. 
    @toopack_pkg_path = params[:pkg_path] || get_latest_toopack_pkg_path
    # Adjust PORTAL_PATH constant as it is required by 'toolpack_defaults.rb'
    PORTAL_PATH.replace "#{@toopack_pkg_path}/web"
    $LOAD_PATH << "#{@toopack_pkg_path}/web/lib"
    $LOAD_PATH << "#{@toopack_pkg_path}/web/components" # for older release like 2.2
    if File.exist? "#{@toopack_pkg_path}/web/lib/toolpack_defaults.rb"
      require "toolpack_defaults"
      # Now that 'toolpack_defaults.rb' is loaded, we have access to toolpack @@defaults varialbles.
      platform = @@defaults[:platform]
      @pre_2_6 = false
    else
      platform = Dir.glob( "#{@toopack_pkg_path}/lib/release/*" ).first.split('/').last
      @pre_2_6 = true
    end
    # Add toolpack DLL path to %PATH% ENV var on windows, and to LD_LIBRARY_PATH on unix.
    env_var = RUBY_PLATFORM =~ /mswin|mingw/ ? 'PATH' : 'LD_LIBRARY_PATH'
    env_delim = RUBY_PLATFORM =~ /mswin|mingw/ ? ';' : ':'
    if ENV[env_var].nil?
      ENV[env_var] = "#{@toopack_pkg_path}/lib/release/#{platform}"
    else
      ENV[env_var] = "#{@toopack_pkg_path}/lib/release/#{platform}" + "#{env_delim}" + ENV[env_var]
    end
  end
  def get_latest_toopack_pkg_path
    toolpack_path = RUBY_PLATFORM =~ /mswin|mingw/ ? 'C:/TelcoBridges/toolpack' : '/lib/tb/toolpack'
    Dir.glob( "#{toolpack_path}/pkg/2.*" ).sort{|x,y| get_toopack_version(x) <=> get_toopack_version(y)}.last
  end
  def get_toopack_version(pkg_path)
    ver,maj,min = /(\d{1,4})\.(\d{1,4})\.(\d{1,4})/.match(pkg_path).captures
    (ver.to_i * 100000) + (maj.to_i * 1000) + min.to_i
  end
end

#
# Class used to return the ERB format for the requested file format
class StatLoggerErbFormat
  def self.csv_format
    # ouputs something like: 
    # date,time,path,item_name,received_bytes_total,sent_bytes_total
    csv_head = <<-END_HEAD
    date,time,path,item_name,<%= @statgrp.fields.join(',') -%>
    END_HEAD
    csv_head = csv_head.gsub(/^\s*/,'')

    # ouputs something like: 
    # "07/26/2013","12:24:37",/host/pm/network,eth0,2044493754,353784475
    csv_body = <<-END_CONTENT
    <% @statgrp.values.each do |path,items| %>
      <% items.each do |item,_fields| %>
        "<%=@statgrp.date%>","<%= @statgrp.time%>",<%=path%>,<%=item%>,<%= @statgrp.fields.collect {|field| _fields[field].to_s}.join(',') %>
      <% end %>
    <% end %>
    END_CONTENT
    csv_body = csv_body.gsub(/^\s*/,'')

    csv_tail = <<-END_TAIL
    END_TAIL
    csv_tail = csv_tail.gsub(/^\s*/,'')

    return [csv_head, csv_body, csv_tail]
  end
  def self.sql_format
    # ouputs something like: 
    # CREATE TABLE [/host/pm/network] (date DATETIME, item_name varchar2(64),[received_bytes_total] varchar2(32),[sent_bytes_total] varchar2(32) );
    sql_head = <<-END_HEAD
    CREATE TABLE [<%=@table_name%>] ( 
        date DATETIME,
        item_name varchar2(64)
      <% @statgrp.path_fields(@render_path).each do |field|%>
        ,[<%=field%>] varchar2(32)
      <%end%>
      );
    END_HEAD
    sql_head = sql_head.gsub(/^\s*/,'')

    # ouputs something like: 
    # INSERT INTO [/host/pm/network] VALUES ( DATETIME('NOW'), 'eth0', '3517361518', '377272798' );
    sql_body = <<-END_CONTENT
    <% @statgrp.values.each do |path,items| %>
      <% items.each do |item,_fields| %>
        INSERT INTO [<%=path%>] VALUES ( DATETIME('now','localtime'),'<%=item%>'
        <% @statgrp.path_fields(path).uniq.each do |field| %> 
          ,'<%= _fields[field].to_s %>'
          <%end%>
          );
      <% end %>
    <% end %>
    END_CONTENT
    sql_body = sql_body.gsub(/^\s*/,'')

    sql_tail = <<-END_TAIL
    END_TAIL
    sql_tail = sql_tail.gsub(/^\s*/,'')

    return [sql_head, sql_body, sql_tail]
  end  

  def self.erb_templates(file)
    case file
    when /.*\.csv/ 
      return csv_format
    when /.*\.xml/
       # todo
    when /.*\.sqlite/
      return sql_format
    else
      puts "File extension unknow for '#{file}', defaulting to CSV"
      return csv_format
    end
  end
end

#
# StatLoggerDeviceFile is based on Ruby Logger
#
class StatLoggerDeviceFile < Logger::LogDevice
   
  # file can be a filename or STDOUT or STDERR
  def initialize(statgrp,params)
    return unless params[:file]
    @statgrp = statgrp
    @file = params[:file]
    @file = STDOUT if (@file == "STDOUT")
    @file = STDERR if (@file == "STDERR")
    rotation_period = params[:rotation_period] || 'daily'
    @slogs_size_max_mb = params[:slogs_size_max_mb] || 10*1024*1024 # 10M
    @cleanup_time = PeriodTimer.new(CLEANUP_PERIOD, 1) # At begining of every CLEANUP_PERIOD
    @head, @body, @tail = StatLoggerErbFormat.erb_templates @file

    $g_log.info "#{@file}: Opening CSV file, rotation_period='#{rotation_period}'"
    super(@file, :shift_age => rotation_period)
  end

  def log_close
    close
  end

  def log_update
    return unless @file
    log_cleanup if @cleanup_time.now? #needs_log_cleanup?
    write_body
  end

  # This method will cleanup old logs, 
  # limiting the size on disk
  def log_cleanup
    return if @file == STDOUT || @file == STDERR
    dirname, filename = File.split(@file)
    extname  = File.extname(filename)
    basename = File.basename(filename, extname)
    logfiles = Dir.glob("#{dirname}/#{basename}.*#{extname}").sort
    kept_size = File.stat(@file).size # Include current file in kept size
    kept_files = []
    # Browsing file the list in reverse order to keep 
    # newest logs up to max size
    logfiles.reverse_each do |file|
      size = File.stat(file).size
      break if kept_size + size >= @slogs_size_max_mb
      kept_size += size
      kept_files << file
    end
    # the difference between files and ketp_files are to be deleted
    del_files = logfiles - kept_files  
    del_files.each { |file| File.unlink(file) rescue nil }
    $g_log.info("#{@file}: log_cleanup -> deleted logfiles #{del_files.inspect}, current size is '#{kept_size/1024}Kb'")
    @cleanup_time.shift
  end

private
  # This method is called each time a new file is created
  def add_log_header(file)
    write_head file
  end

  def write_head file
    file.write(render(@head)) 
    file.write("\n")
  end

  def write_body
    tries ||= FILE_WRITELOCK_RETRY_CNT
    begin
      write(render(@body))
    rescue Errno::EACCES => e
      $g_log.error "#{@file}: write_body CSV is locked, retrying... #{FILE_WRITELOCK_RETRY_CNT-tries+1}/#{FILE_WRITELOCK_RETRY_CNT}"
      sleep 1
      retry unless (tries -= 1).zero?
      $g_log.error "#{@file}: write_body CSV is locked, update data is lost"
    end
  end

  def render erb
    ERB.new(erb,nil,'%<>-').result( binding ).gsub(/^\s*$\n/, '')
  end

  # This method overrides logger.rb shift_log_period
  # to match TB log file name convention
  def shift_log_period(period_end)
    postfix_format = @shift_age =~ /^hourly$/ ? "%Y%m%d.%H%M" : "%Y%m%d"
    postfix = period_end.strftime( postfix_format )
    extname = File.extname(@filename)
    basename = File.basename(@filename, extname)
    age_file = "#{basename}.#{postfix}#{extname}"
    if FileTest.exist?(age_file)
      # try to avoid filename crash caused by Timestamp change.
      idx = 0
      # .99 can be overridden; avoid too much file search with 'loop do'
      while idx < 100
        idx += 1
        age_file = "#{basename}.#{postfix}.#{idx}#{extname}"
        break unless FileTest.exist?(age_file)
      end
    end
    @dev.close rescue nil
    File.rename("#{@filename}", age_file)
    @dev = create_logfile(@filename)
    $g_log.info("#{@file}: shift_log_period -> '#{@filename}' rotated into '#{age_file}'")
    return true
  end
  
  def previous_period_end(now)
    case @shift_age
    when /^hourly$/
      Time.at((now.to_i/3600)*3600)
    when /^daily$/
      eod(now - 1 * SiD)
    when /^weekly$/
      eod(now - ((now.wday + 1) * SiD))
    when /^monthly$/
      eod(now - now.mday * SiD)
    else
      now
    end
  end  
end

#
class StatLoggerDeviceSqlite
  @@databases = {}
 
  def initialize(statgrp,params)
    require 'sqlite3'
    return unless params[:file]
    @params = params
    @file = @params[:file]
    @statgrp = statgrp
    @db = self.class.open_database(@params)
    @db.busy_handler( Proc.new{ |retries| db_busy_handle(retries) } )
    @head, @body, @tail = StatLoggerErbFormat.erb_templates @file
    @rotation_period = @params[:rotation_period] || 'monthly'
    @cleanup_time = PeriodTimer.new(CLEANUP_PERIOD, 1) # At begining of every CLEANUP_PERIOD
    check_db_tables
  end

  def log_close
    self.close @params
  end

  def log_update
    log_cleanup if @cleanup_time.now?
    insert_sql = render @body
    #puts insert_sql
    tries ||= FILE_WRITELOCK_RETRY_CNT
    begin
      @db.execute_batch( insert_sql )
    rescue SQLite3::BusyException => e
      $g_log.error "#{@file}: log_update Database is locked, retrying... #{FILE_WRITELOCK_RETRY_CNT-tries+1}/#{FILE_WRITELOCK_RETRY_CNT}"
      sleep 1
      retry unless (tries -= 1).zero?
      $g_log.error "#{@file}: log_update Database is locked, update data is lost"
    end
  end
  
  def log_cleanup
    previous_period = 
    case @rotation_period
    when 'monthly': @db.execute("SELECT DATETIME('now','-1 month','localtime')") # log entries older than 30 days.
    when 'weekly' : @db.execute("SELECT DATETIME('now','-7 day','localtime')")   # log entries older than 7 days.
    when 'daily'  : @db.execute("SELECT DATETIME('now','-1 day','localtime')")   # log entries older than 1 day.
    when 'hourly' : @db.execute("SELECT DATETIME('now','-1 hour','localtime')")  # log entries older than 1 hour.
    end

    db_tables.each do |table|
      @db.execute( "DELETE FROM [#{table}] WHERE date < '#{previous_period}';" )
      $g_log.info("#{@file}: log_cleanup deleted #{@db.changes} rows from '#{table}' older than '#{previous_period}'")
    end
    @cleanup_time.shift
  end

private
  def self.open_database(params)
    file = params[:file]
    $g_log.info "#{file}: Opening database"
    @@databases[ file ] ||= {}
    @@databases[ file ][ :use_count ] ||= 0

    db = @@databases[ file ][ :db ] || SQLite3::Database.new( file )
    @@databases[ file ][ :db ] = db
    @@databases[ file ][ :use_count ] += 1
    if  @@databases[ file ][ :use_count ] == 1
      # We need to VACUMM sqlite database once in a while, to cleanup internal structure and shrink the file size.
      # This is because DELETE does not free used space in sqlite file.
      # Application startup seems an appropiate time
      size_before = File.stat(file).size
      db.execute( "VACUUM;" )
      size_after = File.stat(file).size
      $g_log.info("#{file}: open_database database was vacuumed, size went from '#{size_before/1024}Kb' to '#{size_after/1024}Kb'")
    end
    return db
  end
  def self.close_database(params)
    file = params[:file]
    @@databases[ file ][ :use_count ] -= 1
    @@databases[ file ][ :db ].close unless @@databases[ file ][ :use_count ] > 0
  end
  def db_tables
    @db.execute( "SELECT name FROM sqlite_master WHERE type='table' ORDER BY name;" )
  end
  def db_busy_handle( retries )
    # Retry handling is done on upper layer, as db_busy_handle 'retries' counter is not working
    # and we want to exit the loop if file gets locked forever... 
    false
  end
  def check_db_tables
    statgrp_tables = @statgrp.paths.collect { |path| [path] }
    missing_tables = statgrp_tables - db_tables
    create_db_tables(missing_tables)
    #statgrp_tables.each { |table| check_db_table_col table } TODO: Check table columns against YML configuration
  end
  def create_db_tables(missing_tables)
    missing_tables.each do |table_name|
      @render_path = table_name.to_s
      @table_name = table_name
      create_sql = render(@head).gsub(/\n/, '')
      #puts create_sql
      $g_log.info "#{@file}: Creating table '#{table_name}' -> '#{create_sql}'"
      @db.execute(create_sql)
    end
  end
  def render erb
    ERB.new(erb,nil,'%<>-').result( binding ).gsub(/^\s*$\n/, '')
  end
end

#
# StatLoggerDevice Factory.  
# We could support other file device such as sqlite
class StatLoggerDevice
  def self.create statgrp,params
  case params[:file]
    when /.*\.sqlite/
      StatLoggerDeviceSqlite.new statgrp,params
    when /.*\.csv/,  /.*\.xml/
      StatLoggerDeviceFile.new statgrp,params
    else
      puts "StatLoggerDevice unknow for '#{params[:file]}', defaulting to FileDevice"
      StatLoggerDeviceFile.new statgrp,params
    end  
  end
end

class StatGroup
  attr_accessor :stat_logger

  def initialize( tbstatus, group_name )
    @group_name = group_name
    @tbstatus = tbstatus
    @group_paths = @tbstatus.group_paths(@group_name)
    @slog_update_interval = @tbstatus.group_param(group_name,'slog_update_interval') || DEFAULT_SLOG_UPDATE_INTERVAL
    @slog_update_interval = SecondInterval.parse( @slog_update_interval )
    @path_status = {}
    @last_update_time = 0
    @update_time = PeriodTimer.new(@slog_update_interval, 0) # Every slog_update_interval, starting now
    @stat_logger = nil
  end

  # Update statistics for the group
  # Build an recursive hash for the values
  def update
    return unless @update_time.now?
    # Cleanup and update each path
    @tbstatus.garbage_collect
    @group_paths.each do |path,fields|
      @path_status[path] = @tbstatus.get(path)
    end

    # Create Ordered hashes with up to date values
    # { "path" => { "item" => { "field" => "value" } } }
    # path: path for the stat (ie "/nap", "/host")
    # item: name of item on the path (ie eth0,eth1)
    # field,value: statistic field name and current value (ie 'sent_bytes_total' => '45033')
    @group_values = OrderedHash.new
    @group_paths.each do |path,fields|
      next if fields.nil?
      item_values = OrderedHash.new
      @path_status[path].each do |name,stats|
        field_values = OrderedHash.new
        fields.each do |field_name|
          field_values[field_name] = get_stats_field_value(stats,field_name).to_s
        end
        item_values[name] = field_values 
      end
     @group_values[path] = item_values
    end
    @last_update_time = Time.now
    @stat_logger.log_update if @stat_logger
    @update_time.shift
  end

  # returns orderd list of all the fields configured in this stat group
  def fields
    field_list = []
    @group_paths.each { |path,fields| (fields||[]).each { |f| field_list << f} }
    field_list
  end 

  def path_fields path
    @group_paths[ path ]
  end

  # returns orderd list of all the paths configured in this stat group
  def paths
    @group_paths.keys
  end

  # Returns recursive hashs object with current statistics values
  # { "path" => { "item" => { "field" => "value" } } }
  # path: path for the stat (ie "/nap", "/host")
  # item: name of item on the path (ie eth0,eth1)
  # field,value: statistic field name and current value (ie 'sent_bytes_total' => '45033')
  def values
    @group_values
  end
  
  # helper method to get the last update date
  def date
    @last_update_time.strftime '%m/%d/%Y'
  end

  # helper method to get the last update time
  def time
    @last_update_time.strftime '%H:%M:%S'
  end

  def get_stats_field_value(stats,field_name)
    field_value = nil
    if field_name =~ /^([a-zA-Z_0-9]*)\.([a-zA-Z_0-9 \(\)]*)$/
      #puts "#{field_name}=#{stats[$1].get_fields[$2]}"
      field_name = $1
      subfield_name = $2
      field_value = case
        when @tbstatus.toolpack_version >= 209000 
          stats[field_name].get_fields[subfield_name] rescue 'N/A' # test stats if supports the requested field, print N/A otherwise
        else
          stats.send(field_name).get_fields[subfield_name] rescue 'N/A' # test stats if supports the requested field, print N/A otherwise
        end
    else
      #puts "#{path}:#{name} -> #{field_name}=#{stats[field_name]}"
      field_value = case
        when @tbstatus.toolpack_version >= 209000 
          stats[field_name] rescue 'N/A' # test stats if supports the requested field, print N/A otherwise
        else
          stats.send(field_name) rescue 'N/A' # test stats if supports the requested field, print N/A otherwise
        end
    end
    field_value
  end
end

class TBStatus < StatusClientWrapper
  @cfg = nil

  def initialize(cfgfile)
    @cfgfile = cfgfile || DEFAULT_YML_CFG_FILE
    load_yml_cfg(@cfgfile)
    init_applog
    params = {}
    params[:gw_port]  = cfg('system_id') unless cfg('system_id').nil?
    params[:pkg_path] = cfg('pkg_path')
    super params
  end

  def cfg(cfg_item)
    @cfg && @cfg['CFG'] && @cfg['CFG'][cfg_item]
  end
  # returns list of defined path fro given group
  def group_paths(group)
    group_param group,'paths'
  end
  def group_param(group,param)
    @cfg['GROUPS'] && @cfg['GROUPS'][group] && @cfg['GROUPS'][group][param]
  end
  # returns list of defined groups
  def cfg_groups
    @cfg['GROUPS'].keys
  end

  # Print all the stats associated with a GROUP name
  def print_stats(group)
    raise "\n\n****** YML config not file found, could not print GROUP stats\n\n" unless @cfg
    case group
    when :all
      groups = cfg_groups
    else
      groups = [group] 
    end
    #puts "groups: #{groups.inspect}"
    groups.each do |group|
      #puts "group:path=>fields: #{group}:#{group_paths(group).inspect}"
      puts "=========== #{group} ================"
      group_paths(group).each do |path,fields|
        #puts "group:path=>fields: #{group}:#{path}=>#{fields}"
        print_path path, fields
      end
    end
  end

  def go_deamon
    raise "\n\n****** YML config not file found, could not go into deamon mode\n\n" unless @cfg
    statgrps = []
    cfg_groups.each do |group_name|
      file = group_param(group_name,'slog_file')
      rotation_period = group_param(group_name,'slog_rotation_period')
      slogs_size_max_mb = group_param(group_name,'slogs_size_max_mb') || DEFAULT_SLOG_SIZE_MAX_MB
      slogs_size_max_mb = FileSize.parse( slogs_size_max_mb )

      statgrp = StatGroup.new(self,group_name)
      if( file ) 
        statgrp.stat_logger = StatLoggerDevice.create( statgrp, 
          {:file => file, :rotation_period => rotation_period, :slogs_size_max_mb => slogs_size_max_mb } )
      end

      statgrps << statgrp
    end
    while 1
      statgrps.each do |statgrp|
        statgrp.update
      end 
      sleep 1
    end
  end  

private
  def init_applog
    logdev = cfg('applog') || STDERR
    logdev = STDERR if (logdev == 'STDERR')
    logdev = STDOUT if (logdev == 'STDOUT')
    $g_log = Logger.new(logdev, 7, 1024*1024)
    $g_log.datetime_format ="%Y-%m-%d %H:%M:%S "
  end

  def load_yml_cfg(file)
    @cfg = parse_yml_file file
  end

  def parse_yml_file file
    return nil unless File.exists? file
    yml_file = YAML.load_file( file )
    #puts yml_file.inspect
    grps = OrderedHash.new
    # create a new Hash object, and transform the array of groups into an ordered hash
    yml_file['GROUPS'].each { |h| grps[h.keys.first.upcase] = h.values.first }
    yml_file['GROUPS'] = grps
    yml_file['GROUPS'].each do |group,params|
      pths = OrderedHash.new
      # create a new Hash object, and transform the array of path into an ordered hash
      params['paths'].each do |h|
        pths[h.keys.first] = h.values.first
        raise "#{file}: Configuration error -> No element defined under path='#{h.keys.first}' for group='#{group}'" if h.values.first.nil? 
      end
      params['paths'] = pths
      
      #oh.each { |key,value| puts "#{key}=>#{value.inspect}" }
    end
    #puts yml_file.inspect
    yml_file
  end
end

cfgfile = DEFAULT_YML_CFG_FILE

if ARGV[0] == nil  : puts(USAGE) ; exit(0) end
if ARGV[0] == "-h" : puts(USAGE) ; exit(0) end
if ARGV[0] == "-f" : cfgfile = ARGV[1]; 2.times { ARGV.shift} end

begin
  tbstatus = TBStatus.new cfgfile
  paths = tbstatus.get_paths.uniq
case 
  when ARGV[0] && paths.include?( ARGV[0].split(':')[0] )
    # print statititics for path given as argument
    # path format example: /nap:SIP_NAP::asr_stats_incoming_struct.last_24h_call_cnt
    args = ARGV[0].split('::')
    tbstatus.print_path( args[0], args[1] ? [args[1]] : nil )
  when ARGV[0] =~ /^[A-Z0-9_]*$/
    # print statititics for specified GROUP
    group = ARGV[0]
    tbstatus.print_stats group
  when ARGV[0] == ":all"
    # print statititics for all defined GROUPS
    tbstatus.print_stats :all
  when ARGV[0] == ":dump"
    # print statititics for all defined GROUPS
    paths.each do |path| 
      next if path =~ /\*/  # We do not want to print path with stars, as we are already browsing the complete tree
      tbstatus.print_path(path)
    end
  when ARGV[0] == "-d"
    # go into deamon mode, and starting logging stats as 
    # defined in yml config file
    tbstatus.go_deamon
  else
    puts "*** ERROR: Path '#{ARGV[0]}' is not available"
    paths.each { |p| puts p }
  end
rescue
  puts $!
  puts $!.backtrace
ensure
  tbstatus.uninitialize rescue nil # rescue to go around bug in 2.2
end
